@@ -1,6 +1,11 @@
 import pandas as pd
 import numpy as np
 import plotly.express as px
+import re
+from multiprocessing import Pool
+import unidecode
+
+import warnings
 
 from pandas.api.types import is_float_dtype, is_numeric_dtype, is_string_dtype
 #from humailib.cloud_tools import GoogleBigQuery, GoogleCloudStorage
@@ -21,14 +26,16 @@ https://humai.sharepoint.com/:w:/s/Platform/ET6ZzN-QgiJIpLdbVMkFz0oBe9_UOhSoau0Z
 humai_table_field_types = {    
     'transactions' : {
         'datetime' : ['Datetime'],
-        'string' : ['Transaction_ID','Branch_ID','Customer_ID','Status','Single_Item_Product_ID'],
-        'float' : ['Total_Gross','Discount','Amount_Paid'],
-        'int' : ['Total_Num_Items_Purchased']
+        'string' : ['Transaction_ID','Branch_ID','Customer_ID','Status','Single_Item_Product_ID','Zip','Item_Types'],
+        'float' : ['Total_Gross','Discount','Amount_Paid','Voucher_Amount'],
+        'int' : ['Total_Num_Items_Purchased'],
+        'bool' : ['Voucher_Used'],
     },
     'items' : {
         'datetime' : ['Datetime'],
-        'string' : ['Transaction_ID','Product_ID','Name'],
-        'float' : ['Total', 'Discount'],
+        'string' : ['Transaction_ID','Product_ID','Name','Item_Type'],
+        'float' : ['Price', 'Discount'],
+        'int' : ['Quantity'],
     },
     'mailings' : {
         'datetime' : ['Send_Datetime'],
@@ -40,28 +47,108 @@ humai_table_field_types = {
         'string' : ['Broadcast_ID','Customer_ID','Activity','Description'],
     },
     'customers' : {
-        'string' : ['Customer_ID','Email','Name','Address'],
+        'string' : ['Customer_ID','Email'],
     },
     'customer_ids' : {
-        'string' : ['Customer_ID','Email','Person']
+        'string' : ['Customer_ID','Personifier','Person']
     },
     'products' : {
-        'string' : ['Name','Product_ID','SKU'],
+        'string' : ['Name','Product_ID','SKU','Item_Type'],
         'float' : ['Price'],
     },
     'customer_characteristics' : {
         'string' : ['Customer_ID']
     },
     'product_characteristics' : {
-        'string' : ['Customer_ID','Product_ID']
+        'string' : ['Product_ID']
     },
     'predictions' : {
         'string' : ['Customer_ID'],
-        'datetime' : ['Date','date_to_email'],
+        'datetime' : ['Date'],
+        #'float' : ['CLV'],
         #'int' : ['Time_Of_Day']
     },
+    'customer_lifetime_value' : {
+        'float' : ['CLV'],
+        'string' : ['Customer_ID'],
+    },
 }
 
+def transactions_check_basic_assumptions(df):
+    
+    assert len(df[df.Amount_Paid < 0]) == 0
+    assert len(df[df.Total_Gross < 0]) == 0
+    assert len(df[df.Discount > 0]) == 0
+    assert len(df) == df.Transaction_ID.nunique()
+    assert len(df[df.Customer_ID == '']) == 0
+    assert len(df[df.Transaction_ID == '']) == 0
+    assert len(df[df.Total_Gross < df.Amount_Paid]) == 0
+    
+    assert len(df[df.Total_Num_Items_Purchased <= 0]) == 0
+    
+    
+def items_check_basic_assumptions(df):
+    
+    assert len(df[df.Product_ID == '']) == 0
+    assert len(df[df.Price < 0]) == 0
+    assert len(df[df.Discount > 0]) == 0
+    assert len(df[df.Quantity <= 0]) == 0
+    
+
+def customers_check_basic_assumptions(df):
+
+    assert len(df) == df.Customer_ID.nunique()
+    assert len(df[df.Customer_ID == '']) == 0
+    
+    
+def products_check_basic_assumptions(df):
+    
+    assert len(df[df.Product_ID == '']) == 0
+    assert len(df) == df.Product_ID.nunique()
+    assert len(df[df.Price < 0]) == 0
+    
+    
+humai_table_checks = {
+    'transactions' : transactions_check_basic_assumptions,
+    'items' : items_check_basic_assumptions,
+    'customers' : customers_check_basic_assumptions,
+    'products' : products_check_basic_assumptions,
+}
+
+
+def check_table(df, table_type, hard_assert=True):
+    
+    print("Checking columns...")
+    for _,cols in humai_table_field_types.get(table_type,[]).items():
+        for c in cols:
+            if c in df:
+                print("  '{}' -> Exists".format(c))
+            else:
+                if hard_assert:
+                    assert c in df, "  '{}' -> Missing!".format(c)
+                else:
+                    warnings.warn(
+                        "  '{}' -> Missing!".format(c),
+                        category=UserWarning, 
+                        stacklevel=1
+                    )
+         
+    if table_type in humai_table_checks:
+        print("Checking basic assumptions...")
+        humai_table_checks[table_type](df)
+                
+                
+def percentage_match_stats(df1, col1, df2, col2):
+    vals1 = set(df1[col1].to_numpy())
+    vals2 = set(df2[col2].to_numpy())
+    
+    print(f"Number of unique '{col1}': {len(vals1)}, unique '{col2}': {len(vals2)}")
+    prop = 100.0 * len(vals1.intersection(vals2)) / len(vals1)
+    print("Proportion of '{}' found in '{}': {:.2f} %".format(col1, col2, prop))
+    
+    diff = list(vals1.difference(vals2))[:5]
+    print("First 5 of the difference set: {}".format(diff))
+    
 
 def convert_table_columns(
     df, 
@@ -94,6 +181,7 @@ def load_table(
     table_type,
     datetime_columns = None,
     datetime_format = '%Y-%m-%d %H:%M:%S', 
+    check_columns = True,
     flush_cache = True,
 ):
     """
@@ -109,8 +197,33 @@ def load_table(
     print("[load_table] Converting table columns...")
     if datetime_columns is not None:
         columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
-    convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)    
+      
     
+    if table_type is not None:
+        
+        do_continue = True
+        if check_columns:
+            for _,columns in humai_table_field_types[table_type].items():
+                for c in columns:
+                    if c not in df:
+                        warnings.warn(
+                            f"[load_table] Column {c} is missing...",
+                            category=UserWarning, 
+                            stacklevel=1
+                        )
+                        do_continue = False
+
+        if not do_continue:
+            return None
+
+        convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)    
+
+        if table_type == 'transactions' and keep_most_recent_trans_id:
+            print("[load_and_merge_tables] Keeping only the most recent Transaction ID, if there's multiples...")
+            print("[load_and_merge_tables]   {:,} rows total.".format(len(df)))
+            keep_most_recent_transaction_id(df)
+            print("[load_and_merge_tables]   {:,} rows left.".format(len(df)))
+
     return df
 
 
@@ -120,6 +233,7 @@ def load_csv(
     table_type,
     datetime_columns = None,
     datetime_format='%Y-%m-%d %H:%M:%S',
+    check_columns=True,
     flush_cache = True
 ):
     """
@@ -135,8 +249,32 @@ def load_csv(
     print("[load_table] Converting table columns...")
     if datetime_columns is not None:
         columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
-    convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
+        
+    if table_type is not None:
+        
+        do_continue = True
+        if check_columns:
+            for _,columns in humai_table_field_types[table_type].items():
+                for c in columns:
+                    if c not in df:
+                        warnings.warn(
+                            f"[load_table] Column {c} is missing...",
+                            category=UserWarning, 
+                            stacklevel=1
+                        )
+                        do_continue = False
+
+        if not do_continue:
+            return None
+        
+        convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
     
+        if table_type == 'transactions' and keep_most_recent_trans_id:
+            print("[load_and_merge_tables] Keeping only the most recent Transaction ID, if there's multiples...")
+            print("[load_and_merge_tables]   {:,} rows total.".format(len(df)))
+            keep_most_recent_transaction_id(df)
+            print("[load_and_merge_tables]   {:,} rows left.".format(len(df)))
+
     return df
 
 
@@ -148,6 +286,8 @@ def load_and_merge_tables(
     table_name = None,
     datetime_columns = None,
     datetime_format = '%Y-%m-%d %H:%M:%S', 
+    check_columns = True,
+    keep_most_recent_trans_id=True,
     flush_cache = True,
 ):
     """
@@ -173,25 +313,76 @@ def load_and_merge_tables(
             if df is None:
                 df = df_tmp.copy()
             else:
-                df = df.append(df_tmp)
+                df = df.append(df_tmp, sort=True, ignore_index=True)
 
     if df is None:
         return None
     
-    print("[load_and_merge_tables] {:,} rows total.".format(len(df)))
+    do_continue = True
+    if check_columns:
+        for _,columns in humai_table_field_types[table_type].items():
+            for c in columns:
+                if c not in df:
+                    warnings.warn(
+                        f"[load_and_merge_tables] Column {c} is missing...",
+                        category=UserWarning, 
+                        stacklevel=1
+                    )
+                    do_continue = False
+                    
+    if not do_continue:
+        return None
+    
     print("[load_and_merge_tables] Dropping duplicate rows...")
+    print("[load_and_merge_tables]   {:,} rows total.".format(len(df)))
     df.drop_duplicates(inplace=True)
-    print("[load_and_merge_tables] {:,} rows left.".format(len(df)))
+    print("[load_and_merge_tables]   {:,} rows left.".format(len(df)))
     
     print("[load_and_merge_tables] Converting table columns...")
     if datetime_columns is not None:
         columns_as_datetime(df, columns=datetime_columns, datetime_format=datetime_format)
     convert_table_columns(df, table_type=table_type, datetime_format=datetime_format)
     
+    if table_type == 'transactions' and keep_most_recent_trans_id:
+        print("[load_and_merge_tables] Keeping only the most recent Transaction ID, if there's multiples...")
+        print("[load_and_merge_tables]   {:,} rows total.".format(len(df)))
+        keep_most_recent_transaction_id(df)
+        print("[load_and_merge_tables]   {:,} rows left.".format(len(df)))
+    
     return df
 
 
-def merge_dataframes(dataframes, on, overwrite_cols=False):
+def upload_and_append_table_unique(
+    gbq,
+    dataset_table,
+    df,
+    unique_subset,
+    datetime_columns=[],
+    table_type=None,
+    check_columns=False,
+    flush_cache=True,
+):
+    df_existing = load_table(
+        gbq,
+        dataset_table,
+        table_type=table_type,
+        datetime_columns=datetime_columns,
+        check_columns=check_columns,
+        flush_cache=flush_cache,
+    )
+    
+    print(f"Size of existing table before appending: {len(df_existing)}")
+    df_existing = df_existing.append(df, sort=True, ignore_index=True)
+    df_existing.drop_duplicates(subset=unique_subset, keep='first', inplace=True)
+    print(f"Size of table after appending and removing duplicates: {len(df_existing)}")
+    
+    gbq.upload_and_replace_table(
+        df_existing,
+        dataset_table,
+    )
+    
+
+def merge_dataframes(dataframes, on, how='left', overwrite_cols=False):
     
     df_out = None
     for df in dataframes:
@@ -207,6 +398,9 @@ def merge_dataframes(dataframes, on, overwrite_cols=False):
                     copy=False,
                 )
             else:
+                for c in df.columns:
+                    if c in df_out:
+                        del df_out[c]
                 df_out = df_out.merge(
                     df,
                     on=on,
@@ -216,16 +410,69 @@ def merge_dataframes(dataframes, on, overwrite_cols=False):
                 
     return df_out
 
+
+def keep_most_recent_transaction_id(df, inplace=True):
+    keep_most_recent_id(df, id_column='Transaction_ID', inplace=inplace)
+    
+    
+def keep_most_recent_id(df, id_column='Transaction_ID', inplace=True):
+    if not inplace:
+        return df.rename_axis('_index').sort_values(by=[id_column,'Datetime','_index']).drop_duplicates(subset=id_column, keep='last')
+
+    df.rename_axis('_index').sort_values(by=[id_column,'Datetime','_index'], inplace=True)
+    df.drop_duplicates(subset=id_column, keep='last', inplace=True)
+    
     
 def force_string(value):
     return 'S' + str(value)#'S' + str(value)
 
 
+def get_papermill_param(param, type_str='str', vis=None):
+    
+    if vis is None:
+        vis = globals()
+        
+    var_lookup = {
+        'int' : 0,
+        'float' : 0.0,
+        'list' : [],
+        'dict' : {},
+        'bool' : False,
+        'str' : '',
+    }
+
+    value = var_lookup.get(type_str, None)
+    if param in vis:
+        
+        value = vis[param] if vis[param] is not None else var_lookup.get(type_str, None)
+        print("PAPERMILL -> Set '{}' to {}".format(
+            param, value
+        ))
+            
+    return value
+
+
+def gbq_friendly_column_name(column):
+    
+    col_name = column
+    if str.isdigit(col_name[0]):
+        col_name = '_' + col_name
+        
+    return unidecode.unidecode(re.sub(r'[^\w]', '_', col_name).replace('-','_')).replace('+','plus').replace('-','to')
+
+
+def bq_friendly_column_names(columns):
+
+    return [gbq_friendly_column_name(c) for c in columns]
+    
+
 def column_stats(df, col_name, n=-1):
     """
     Print out column statistics.
     """
     
+    print("-------------------------")
+    print("COLUMN_STATS for '{}'".format(col_name))
     isna = df[col_name].isna().sum()
     print("NaNs: {:,} out of {:,} ({:.2f}%)".format(isna, len(df), 100.0*isna/len(df)))
     print("== Describe ==")
@@ -266,6 +513,9 @@ def columns_as_str(df, columns, uppercase=False, drop_empty=False, replace_na_va
     
     if columns is None:
         columns = df.columns
+        
+    if isinstance(columns, str):
+        columns = [columns]
 
     if not isinstance(columns, list):
         raise Exception("Expecting columns to be a list.")
@@ -297,6 +547,9 @@ def columns_as_datetime(df, columns, datetime_format='%Y-%m-%d %H:%M:%S'):
     
     if columns is None:
         columns = df.columns
+        
+    if isinstance(columns, str):
+        columns = [columns]
 
     if not isinstance(columns, list):
         raise Exception("Expecting columns to be a list.")
@@ -319,11 +572,16 @@ def columns_as_float(df, columns, na_replace_value=None, drop_na=True):
     if columns is None:
         columns = df.columns
         
+    if isinstance(columns, str):
+        columns = [columns]
+        
     if not isinstance(columns, list):
         raise Exception("Expecting columns to be a list.")
 
     if na_replace_value is None:
         na_replace_value = np.nan
+    else:
+        drop_na = False
         
     for c in columns:
         if c in df and not is_float_dtype(df[c]):
@@ -337,9 +595,15 @@ def columns_as_float(df, columns, na_replace_value=None, drop_na=True):
                 df.dropna(subset=[c], how='any', inplace=True)
             else:
                 df.loc[:,c].fillna(na_replace_value, inplace=True)
-                
+
             df.loc[:,c] = df[c].astype('float64', skipna=False)
         else:
+            if is_float_dtype(df[c]):
+                if drop_na:
+                    df.dropna(subset=[c], how='any', inplace=True)
+                else:
+                    df.loc[:,c].fillna(na_replace_value, inplace=True)
+                
             if c not in df:
                 print("Column '{}' doesn't exist.".format(c))
 
@@ -353,6 +617,15 @@ def columns_as_int(df, columns, na_replace_value=None, drop_na=True):
     Note: This is done inplace.
     """
     
+    if columns is None:
+        columns = df.columns
+        
+    if isinstance(columns, str):
+        columns = [columns]
+        
+    if not isinstance(columns, list):
+        raise Exception("Expecting columns to be a list.")
+        
     if not drop_na and na_replace_value is None:
         na_replace_value = -1.0
 
@@ -365,10 +638,10 @@ def columns_as_int(df, columns, na_replace_value=None, drop_na=True):
             print("Column '{}' doesn't exist.".format(c))
             
             
-def agg_first(series):
+def agg_first(series, **kwargs):
     return series.iloc[0]
 
-def agg_last(series):
+def agg_last(series, **kwargs):
     return series.iloc[-1]
 
 def agg_nth(series, **kwargs):
@@ -378,9 +651,21 @@ def agg_nth(series, **kwargs):
     
     return series.iloc[n]
 
-def agg_list(series):
+def agg_list(series, **kwargs):
     return "{}".format(list(series.to_numpy()))
 
+
+def agg_sequence(series, **kwargs):
+    
+    d = kwargs['delimiter']
+    s = d.join(item for item in series.to_numpy().astype(str))
+    
+    if kwargs['do_sorting']:
+        return d.join(item for item in np.sort( list(set(s.split(d))) ))
+    
+    return d.join(item for item in list(set(s.split(d))))
+
+
 def agg_umode(series):
     modes = series.mode()
     return modes.iloc[0]
@@ -398,8 +683,11 @@ def aggregate_columns(df, agg_type, agg_column, columns, copy_agg_column=False,
         print("[aggregate_columns] '{}' does not exist!!".format(agg_column))
         return None
     
+    if len(columns) == 0:
+        return None
+    
     for c in columns:
-        if c not in columns:
+        if c not in df:
             print("[aggregate_columns] '{}' does not exist!!".format(c))
     
     aggregators = {
@@ -408,11 +696,13 @@ def aggregate_columns(df, agg_type, agg_column, columns, copy_agg_column=False,
         'min'  : pd.DataFrame.min,
         'max'  : pd.DataFrame.max,
         'std'  : pd.DataFrame.std,
+        'any'  : pd.DataFrame.any,
         'mode' : agg_umode,
         'first': agg_first,
         'last' : agg_last,
         'nth'  : agg_nth,
         'list' : agg_list,
+        'seq'  : agg_sequence,
     }
     
     if agg_type in aggregators:
@@ -421,9 +711,12 @@ def aggregate_columns(df, agg_type, agg_column, columns, copy_agg_column=False,
         raise NameError('[aggregate_columns] Need to add {} to transform function'.format(agg_type))
         
     cols = [c for c in columns if c in df]
+    assert len(cols) > 0, f"[aggregate_columns] The columns [{columns}] to aggregate {agg_type} over is the empty array, which means we will aggregate over ALL columns. Do you mean to do this?"
     return df.groupby(by=agg_column)[cols].agg(func, **kwargs)
 
 
+
+
 def transactions_cleanup(df, datetime_column, tid_column = 'Transaction_ID', cid_column = 'Customer_ID', datetime_format='%Y-%m-%d %H:%M:%S'):
     
     """
@@ -468,7 +761,7 @@ def email_ensure_zero_or_one_activity_per_broadcast(
 
     """
     Ensure each email sent to each customer has only one event: either a delivery, an open, a click, or a transaction,
-    in that order. Keep only the first (in time) occurrence of it.
+    in that order. Keep only the first (in time) occurrence of it, in case of duplicate events.
 
     Note: This is done inplace.
     """
@@ -605,21 +898,72 @@ def transaction_stats(
     return fig, df_trans_hist
 
 
-def keep_subset_customers(df, cid_column, k):
+def keep_subset_customers(dataframes, cid_column, k, seed=1):
+    """
+    Keep a subset of customers among all dataframes.
+    
+    Input:
+    dataframes - Either a single dataframe or a list of dataframes.
+    
+    Is not done inplace.
+    """
     import random
     
+    random.seed(seed)
+    
+    dataframes_in = dataframes
+    if isinstance(dataframes, list):
+        df = dataframes[0]
+    else:
+        df = dataframes
+        dataframes_in = [dataframes]
+    
     print("{:,} unique customers".format(df[cid_column].nunique()))
     
     cids = list(df[cid_column].unique())
     to_keep_cids = random.sample(cids, k=k)
     to_keep_cids_map = {cid:1 for cid in to_keep_cids}
 
-    df.loc[:,'keep'] = df[cid_column].apply(lambda x: to_keep_cids_map.get(x,0))
-    df_keep = df.drop(index=df[df.keep == 0].index)
+    dataframes_out = []
+    for df in dataframes_in:
+        df.loc[:,'keep'] = df[cid_column].apply(lambda x: to_keep_cids_map.get(x,0))
+        dataframes_out.append( df.drop(index=df[df.keep == 0].index))
     
-    print("{:,} unique customers after sampling".format(df_keep[cid_column].nunique()))
+    print("{:,} unique customers after sampling".format(dataframes_out[0][cid_column].nunique()))
     
-    return df_keep
+    if isinstance(dataframes, list):
+        return dataframes_out
+    
+    return dataframes_out[0]
+
+
+def filter_dataframe(df, match_filters=None, contains_filters=None):
+    #
+    # Perform any filtering
+    #
+    if match_filters is not None:
+        for c,v in match_filters.items():
+            if c in df:
+                filter_vals = '|'.join(v)
+
+                column_stats(df, c)
+                print(f"Before 'match' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
+                df = df.loc[df[c].str.match(filter_vals),:].copy()
+                print(f"After 'match' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
+                column_stats(df, c)
+                
+    if contains_filters is not None:
+        for c,v in contains_filters.items():
+            if c in df:
+                filter_vals = '|'.join(v)
+
+                column_stats(df, c)
+                print(f"Before 'contains' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
+                df = df.loc[df[c].str.contains(filter_vals),:].copy()
+                print(f"After 'contains' filtering column '{c}' -> '{filter_vals}': n={len(df)}")
+                column_stats(df, c)
+
+    return df
 
 
 def mann_whitney_test(dfA, dfB, column, group_A_name='grp_A', group_B_name='grp_B', scale = 1.0, unit='', verbose=False):
@@ -670,19 +1014,33 @@ def mann_whitney_test(dfA, dfB, column, group_A_name='grp_A', group_B_name='grp_
             p
         ))
         
+        #df_out = pd.DataFrame(data, columns=[
+        #    f'driver',
+        #    #f'sum_{group_A_name}',
+        #    f'mean_{group_A_name}',
+        #    f'median_{group_A_name}',
+        #    f'size_{group_A_name}',
+        #    
+        #    #f'sum_{group_B_name}',
+        #    f'mean_{group_B_name}',
+        #    f'median_{group_B_name}',
+        #    f'size_{group_B_name}',
+        #    
+        #    f'mann_whitney_p_value',
+        #])
         df_out = pd.DataFrame(data, columns=[
-            f'driver',
+            'column',
             #f'sum_{group_A_name}',
-            f'mean_{group_A_name}',
-            f'median_{group_A_name}',
-            f'size_{group_A_name}',
+            '{}_mean'.format(group_A_name),
+            '{}_median'.format(group_A_name),
+            '{}_size'.format(group_A_name),
             
             #f'sum_{group_B_name}',
-            f'mean_{group_B_name}',
-            f'median_{group_B_name}',
-            f'size_{group_B_name}',
+            '{}_mean'.format(group_B_name),
+            '{}_median'.format(group_B_name),
+            '{}_size'.format(group_B_name),
             
-            f'mann_whitney_p_value',
+            'mann_whitney_p_value',
         ])
         
         #print(df_out.head())
@@ -694,3 +1052,22 @@ def mann_whitney_test(dfA, dfB, column, group_A_name='grp_A', group_B_name='grp_
         return None, np.nan
     
     return None, np.nan
+
+
+def parallelise_dataframe(df, func, on_cids=False, n_cores=2):
+    if on_cids:
+        print("[parallelise_data_frame_on_cids] Assigning Customer_IDs to cores...")
+        cid2batch = {cid:(i % n_cores) for i,cid in enumerate(df.Customer_ID.unique())}
+        df.loc[:,'core'] = df.Customer_ID.apply(lambda cid: cid2batch.get(cid, 0))
+        df_split = [df[df.core == i] for i in range(n_cores)]
+    else:
+        df_split = np.array_split(df, n_cores)
+
+    print(f"Executing on {n_cores} cores...")
+    pool = Pool(n_cores)
+    df = pd.concat(pool.map(func, df_split))
+
+    pool.close()
+    pool.join()
+
+    return df
